
import pkg_resources

_fedora_version = pkg_resources.get_distribution('fedora_messaging').version
_pika_version = pkg_resources.get_distribution('pika').version

#: The default client properties reported to the AMQP broker in the "start-ok"
#: method of the connection negotiation. This allows the broker administrators
#: to easily identify what a connection is being used for and the client's
#: capabilities.
DEFAULT_CLIENT_PROPERTIES = {
    'app': 'Unknown',
    'product': 'Fedora Messaging with Pika',
    'information': 'https://fedora-messaging.readthedocs.io/en/stable/',
    'version': 'fedora_messaging-{} with pika-{}'.format(_fedora_version, _pika_version),
}



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
from __future__ import absolute_import, unicode_literals

import json

import jsonschema
import pika
from pika.adapters import twisted_connection
from twisted.internet import interfaces, reactor, defer, protocol
from twisted.logger import Logger
from zope.interface import implementer

from .. import config
from ..message import get_class, Message
from ..exceptions import Nack, Drop, HaltConsumer


log = Logger()


@implementer(interfaces.IPushProducer)
class MessageProducer:
    """
    Twisted `Producer`_ that consumes messages from the AMQP server and passes
    them to a callback.

    .. Producer: https://twistedmatrix.com/documents/current/core/howto/producers.html

    Attributes:
        producing (Deferred): Deferred that will fire when production
            has stopped. The callback value is undefined for now. It is
            ``None`` when the production has not started yet.
    """

    def __init__(self, callback, bindings=None):
        """
        Initializes the producer.

        Args:
            callback (callable): the function that will be called with the
                received message as only argument.
            bindings (dict): the exchanges, queues and bindings to set. If
                ``None``, the configuration file will be used.
        """
        self._consumer_callback = callback
        self._bindings = bindings or config.conf["bindings"]
        self._parameters = pika.URLParameters(config.conf['amqp_url'])
        if self._parameters.client_properties is None:
            self._parameters.client_properties = config.conf['client_properties']
        self._connection = None
        self._channel = None
        self._running = False
        self.producing = None

    @defer.inlineCallbacks
    def resumeProducing(self):
        """
        Starts or resumes the connection to the server.

        This method starts receiving messages from the server, they will be
        passed to the consumer callback.

        If necessary, the connection is established, the channel is open and
        the configured queues and exchanges are declared and bound.

        Returns:
            Deferred: fired when the production is ready to start
        """
        # Connect to server
        if self._connection is None:
            cc = protocol.ClientCreator(
                reactor,
                twisted_connection.TwistedProtocolConnection,
                self._parameters
            )
            if self._parameters.ssl:
                # TODO: SSL certs?
                connectMethod = cc.connectSSL
            else:
                connectMethod = cc.connectTCP
            self._connection = yield connectMethod(self._parameters.host, self._parameters.port)
            yield self._connection.ready
            log.debug("Connected to AMQP server")
        # Create channel
        if self._channel is None:
            self._channel = yield self._connection.channel()
            log.debug("AMQP channel created")
            yield self._channel.basic_qos(prefetch_count=0, prefetch_size=0)
        # Declare exchanges and queues
        queues = set()
        for binding in self._bindings:
            yield self._channel.exchange_declare(
                exchange=binding['exchange'],
                exchange_type='topic', durable=True)
            result = yield self._channel.queue_declare(
                queue=binding['queue_name'],
                durable=True,
                auto_delete=binding.get("queue_auto_delete", False),
                arguments=binding.get('queue_arguments'),
            )
            queue_name = result.method.queue
            yield self._channel.queue_bind(
                queue=queue_name,
                exchange=binding['exchange'],
                routing_key=binding['routing_key'],
            )
            queues.add(queue_name)
        log.debug("AMQP bindings declared")
        # Start consuming
        queue_objects = []
        for queue_name in queues:
            queue_object, consumer_tag = yield self._channel.basic_consume(
                queue=queue_name)
            queue_objects.append(queue_object)
        self._running = True
        self.producing = defer.DeferredList([
            self._read(qo) for qo in queue_objects
        ])
        log.info('AMQP consumer is ready')

    @defer.inlineCallbacks
    def pauseProducing(self):
        """
        Pause the reception of messages. Does not disconnect from the server.

        Message reception can be resumed with :meth:`resumeProducing`.

        Returns:
            Deferred: fired when the production is paused.
        """
        if self._channel is None:
            return
        for tag in self._channel.consumer_tags:
            yield self._channel.basic_cancel(tag)
        self._running = True
        log.debug("Producing paused.")

    @defer.inlineCallbacks
    def stopProducing(self):
        """
        Stop producing messages and disconnect from the server.

        Returns:
            Deferred: fired when the production is stopped.
        """
        if not self._running:
            return
        if self._channel:
            log.info('Halting {tag} consumer sessions', tag=self._channel.consumer_tags)
        self._running = False
        yield self._connection.close()
        self._channel = None
        self._connection = None
        self.producing = None

    @defer.inlineCallbacks
    def _read(self, queue_object):
        while self._running:
            try:
                channel, delivery_frame, properties, body = yield queue_object.get()
            except pika.exceptions.ChannelClosed as e:
                log.error(repr(e))
                break
            except Exception:
                log.failure("Failed getting the next message in the queue. Stopping.")
                break
            if body:
                yield self._on_message(channel, delivery_frame, properties, body)
        if self._running:
            # We broke the loop, something went wrong
            self.stopProducing()

    @defer.inlineCallbacks
    def _on_message(self, channel, delivery_frame, properties, body):
        """
        Callback when a message is received from the server.

        This method wraps a user-registered callback for message delivery. It
        decodes the message body, determines the message schema to validate the
        message with, and validates the message before passing it on to the user
        callback.

        This also handles acking, nacking, and rejecting messages based on
        exceptions raised by the consumer callback. For detailed documentation
        on the user-provided callback, see the user guide on consuming.

        Args:
            channel (pika.channel.Channel): The channel from which the message
                was received.
            delivery_frame (pika.spec.Deliver): The delivery frame which includes
                details about the message like content encoding and its delivery
                tag.
            properties (pika.spec.BasicProperties): The message properties like
                the message headers.
            body (bytes): The message payload.

        Returns:
            Deferred: fired when the message has been handled.
        """
        log.debug('Message arrived with delivery tag {tag}',
                  tag=delivery_frame.delivery_tag)
        if properties.content_encoding is None:
            log.error('Message arrived without a content encoding')
            properties.content_encoding = 'utf-8'
        try:
            body = body.decode(properties.content_encoding)
        except UnicodeDecodeError:
            log.error(
                'Unable to decode message body {body!r} with {encoding!r} '
                'content encoding',
                body=body,
                encoding=delivery_frame.content_encoding
            )
        try:
            body = json.loads(body)
        except ValueError:
            log.failure('Failed to load message body {body!r}', body)
        try:
            MessageClass = get_class(properties.headers['fedora_messaging_schema'])
        except TypeError:
            log.warn(
                'Message (body={body!r}) arrived without headers. '
                'A publisher is misbehaving!', body=body
            )
            MessageClass = Message
        except KeyError:
            log.warn(
                'Message (headers={headers!r}, body={body!r}) arrived without '
                'a schema header. A publisher is misbehaving!',
                headers=properties.headers, body=body
            )
            MessageClass = Message

        message = MessageClass(
            body=body, headers=properties.headers, topic=delivery_frame.routing_key)
        try:
            message.validate()
            log.debug('Successfully validated message {msg!r}', msg=message)
        except jsonschema.exceptions.ValidationError:
            log.failure('Message validation of {msg!r} failed', message)
        try:
            log.info('Consuming message from topic {topic!r} (id {msgid})',
                     topic=message.topic,
                     msgid=properties.message_id)
            yield defer.maybeDeferred(self._consumer_callback, message)
            yield channel.basic_ack(delivery_tag=delivery_frame.delivery_tag)
        except Nack:
            log.info('Returning message id {msgid} to the queue',
                     msgid=properties.message_id)
            yield channel.basic_nack(delivery_tag=delivery_frame.delivery_tag, requeue=True)
        except Drop:
            log.info('Dropping message id {msgid}', msgid=properties.message_id)
            yield channel.basic_nack(delivery_tag=delivery_frame.delivery_tag, requeue=False)
        except HaltConsumer:
            log.warn('Consumer indicated it wishes consumption to halt, shutting down')
            yield self.stopProducing()
        except Exception:
            log.failure("Received unexpected exception from consumer callback")
            yield channel.basic_nack(delivery_tag=0, multiple=True, requeue=True)
            yield self.stopProducing()



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
The ``fedora-messaging`` `Click`_ CLI.

.. _Click: http://click.pocoo.org/
"""
from __future__ import absolute_import

import importlib
import logging
import logging.config

import click

from . import config, api

_log = logging.getLogger(__name__)


@click.group()
@click.option('--conf', envvar='FEDORA_MESSAGING_CONF')
def cli(conf):
    """The fedora-messaging command line interface."""
    if conf:
        try:
            config.conf.load_config(filename=conf)
        except ValueError as e:
            raise click.exceptions.BadParameter(e)


@cli.command()
@click.option('--app-name')
@click.option('--callback')
@click.option('--routing-key')
@click.option('--queue-name')
@click.option('--exchange')
@click.option('--amqp-url')
def consume(amqp_url, exchange, queue_name, routing_key, callback, app_name):

    amqp_url = amqp_url or config.conf['amqp_url']
    if exchange and queue_name and routing_key:
        bindings = [{
            'exchange': exchange,
            'queue_name': queue_name,
            'routing_key': routing_key
        }]
    elif not exchange and not queue_name and not routing_key:
        bindings = config.conf['bindings']
    else:
        raise click.ClickException(
            'You must define all three of exchange, queue_name and'
            ' routing_key, or none of them to use the configuration')
    if not bindings:
        raise click.ClickException(
            'No bindings are defined in the configuration file'
            ' and none were provided as arguments!')

    callback_path = callback or config.conf['callback']
    if not callback_path:
        raise click.ClickException('"callback" must be the Python path to a'
                                   ' callable to consume')
    try:
        module, cls = callback_path.strip().split(':')
    except ValueError as e:
        raise click.ClickException('Unable to parse the callback path ({}); the '
                                   'expected format is "my_package.module:'
                                   'callable_object"'.format(str(e)))
    try:
        module = importlib.import_module(module)
    except ImportError as e:
        raise click.ClickException('Failed to import the callback module ({})'.format(str(e)))

    try:
        callback = getattr(module, cls)
    except AttributeError as e:
        raise click.ClickException(
            'Unable to import {} ({}); is the package installed? The python path should '
            'be in the format "my_package.module:callable_object"'.format(
                callback_path, str(e)))

    if app_name:
        config.conf['client_properties']['app'] = app_name

    _log.info('Starting consumer with %s callback', callback_path)
    return api.consume(callback, bindings)



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
fedora-messaging can be configured with the
``/etc/fedora-messaging/config.toml`` file or by setting the
``FEDORA_MESSAGING_CONF`` environment variable to the path of the configuration
file.

Each configuration option has a default value.

Generic Options
===============

These options apply to both consumers and publishers.

.. _conf-amqp-url:

amqp_url
--------
The AMQP broker to connect to. This URL should be in the format described by
the :class:`pika.connection.URLParameters` documentation. This defaults to
``amqp://``.

.. _conf-client-properties:

client_properties
-----------------
A dictionary that describes the client to the AMQP broker. This makes it easy
to identify the application using a connection. The dictionary can contain
arbitrary string keys and values. The default is::

    {
        'app': 'Unknown',
        'product': 'Fedora Messaging with Pika',
        'information': 'https://fedora-messaging.readthedocs.io/en/stable/',
        'version': 'fedora_messaging-<version> with pika-<version>',
    }

It's recommended that applications only change the ``app`` key in the default
set of keys.

.. _conf-exchanges:

exchanges
---------
A dictionary of exchanges that should be present in the broker. Each key should
be an exchange name, and the value should be a dictionary with the exchange's
configuration.  Options are:

* ``type`` - the type of exchange to create.

* ``durable`` - whether or not the exchange should survive a broker restart.

* ``auto_delete`` - whether or not the exchange should be deleted once no queues
  are bound to it.

* ``arguments`` - dictionary of arbitrary keyword arguments for the exchange,
  which depends on the broker in use and its extensions.

For example::

    {
        'my_exchange': {
            'type': 'fanout',
            'durable': True,
            'auto_delete': False,
            'arguments': {},
        },
    }

The default is to ensure the 'amq.topic' topic exchange exists which should be
sufficient for most use cases.

.. _conf-log-config:

log_config
----------
A dictionary describing the logging configuration to use, in a format accepted
by :func:`logging.config.dictConfig`.


Publisher Options
=================

The following configuration options are publisher-related.

.. _conf-publish-exchange:

publish_exchange
----------------
A string that identifies the exchange to publish to. The default is
``amq.topic``.


.. _sub-config:

Consumer Options
================

The following configuration options are consumer-related.

.. _conf-queues:

queues
------
A dictionary of queues that should be present in the broker. Each key should be
a queue name, and the value should be a dictionary with the queue's configuration.
Options are:

* ``durable`` - whether or not the queue should survive a broker restart.

* ``auto_delete`` - whether or not the queue should be deleted once the
  consumer disconnects.

* ``exclusive`` - whether or not the queue is exclusive to the current
  connection.

* ``arguments`` - dictionary of arbitrary keyword arguments for the queue, which
  depends on the broker in use and its extensions.

For example::

    {
        'my_queue': {
            'durable': True,
            'auto_delete': True,
            'exclusive': False,
            'arguments': {},
        },
    }


.. _conf-bindings:

bindings
--------
A list of dictionaries that define queue bindings to exchanges that consumers
will subscribe to. The ``queue`` key is the queue's name. The ``exchange`` key
should be the exchange name and the ``routing_keys`` key should be a list of
routing keys. For example::

    [
        {
            'queue': 'name',
            'exchange': 'amq.topic',
            'routing_keys': ['topic1', 'topic2.#'],
        },
    ]

This would create two bindings for the ``my_queue`` queue, both to the
``amq.topic`` exchange. Consumers will consume from both queues.

callback
--------
The Python path of the callback. This should be in the format
``<module>:<object>``. For example, if the callback was called "my_callback"
and was located in the "my_module" module of the "my_package" package, the path
would be defined as ``my_package.my_module:my_callback``. The default is None.

Consult the :ref:`consumers` documentation for details on implementing a
callback.

consumer_config
---------------
A dictionary for the consumer to use as configuration. The consumer should
access this key in its callback for any configuration it needs. Defaults to
an empty dictionary.

qos
---
The quality of service settings to use for consumers. This setting is a
dictionary with two keys. ``prefetch_count`` specifies the number of messages
to pre-fetch from the server. Pre-fetching messages improves performance by
reducing the amount of back-and-forth between client and server. The downside
is if the consumer encounters an unexpected problem, messages won't be returned
to the queue and sent to a different consumer until the consumer times out.
``prefetch_size`` limits the size of pre-fetched messages (in bytes), with 0
meaning there is no limit. The default settings are::

    {
        'prefetch_count': 10,
        'prefetch_size': 0,
    }
"""
from __future__ import unicode_literals

import logging
import logging.config
import os
import sys
import uuid

import pkg_resources
import pytoml


_log = logging.getLogger(__name__)

_fedora_version = pkg_resources.get_distribution('fedora_messaging').version
_pika_version = pkg_resources.get_distribution('pika').version

# A default, auto-deleted queue for consumers
_default_queue_name = str(uuid.uuid4())

#: A dictionary of application configuration defaults.
DEFAULTS = dict(
    amqp_url='amqp://',
    #: The default client properties reported to the AMQP broker in the "start-ok"
    #: method of the connection negotiation. This allows the broker administrators
    #: to easily identify what a connection is being used for and the client's
    #: capabilities.
    client_properties={
        'app': 'Unknown',
        'product': 'Fedora Messaging with Pika',
        'information': 'https://fedora-messaging.readthedocs.io/en/stable/',
        'version': 'fedora_messaging-{} with pika-{}'.format(_fedora_version, _pika_version),
    },
    publish_exchange='amq.topic',
    exchanges={
        'amq.topic': {
            'type': 'topic',
            'durable': True,
            'auto_delete': False,
            'arguments': {},
        },
    },
    queues={
        _default_queue_name: {
            'durable': False,
            'auto_delete': True,
            'exclusive': False,
            'arguments': {},
        },
    },
    bindings=[
        {
            'queue': _default_queue_name,
            'exchange': 'amq.topic',
            'routing_keys': ['#'],
        },
    ],
    qos={
        'prefetch_size': 0,
        'prefetch_count': 10,
    },
    callback=None,
    consumer_config={},
    log_config={
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'simple': {
                'format': '[%(name)s %(levelname)s] %(message)s',
            },
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'simple',
                'stream': 'ext://sys.stdout',
            }
        },
        'loggers': {
            'fedora_messaging': {
                'level': 'INFO',
                'propagate': False,
                'handlers': ['console'],
            },
        },
        # The root logger configuration; this is a catch-all configuration
        # that applies to all log messages not handled by a different logger
        'root': {
            'level': 'WARNING',
            'handlers': ['console'],
        },
    },
)

# Start with a basic logging configuration, which will be replaced by any user-
# specified logging configuration when the configuration is loaded.
logging.config.dictConfig(DEFAULTS['log_config'])


def load(filename=None):
    """
    Load application configuration from a file and merge it with the default
    configuration.

    If the ``FEDORA_MESSAGING_CONF`` environment variable is set to a
    filesystem path, the configuration will be loaded from that location.
    Otherwise, the path defaults to ``/etc/fedora-messaging/config.toml``.
    """
    config = DEFAULTS.copy()

    if filename:
        config_path = filename
    elif 'FEDORA_MESSAGING_CONF' in os.environ:
        config_path = os.environ['FEDORA_MESSAGING_CONF']
    else:
        config_path = '/etc/fedora-messaging/config.toml'

    if os.path.exists(config_path):
        _log.info('Loading configuration from {}'.format(config_path))
        with open(config_path) as fd:
            try:
                file_config = pytoml.loads(fd.read())
                for key in file_config:
                    config[key.lower()] = file_config[key]
            except pytoml.core.TomlError as e:
                _log.error('Failed to parse {}: {}'.format(config_path, str(e)))
                sys.exit(1)
    else:
        _log.info('The configuration file, {}, does not exist.'.format(config_path))

    return config


class LazyConfig(dict):
    """This class lazy-loads the configuration file."""
    loaded = False

    def __getitem__(self, *args, **kw):
        if not self.loaded:
            self.load_config()
        return super(LazyConfig, self).__getitem__(*args, **kw)

    def get(self, *args, **kw):
        if not self.loaded:
            self.load_config()
        return super(LazyConfig, self).get(*args, **kw)

    def pop(self, *args, **kw):
        if not self.loaded:
            self.load_config()
        return super(LazyConfig, self).pop(*args, **kw)

    def copy(self, *args, **kw):
        if not self.loaded:
            self.load_config()
        return super(LazyConfig, self).copy(*args, **kw)

    def update(self, *args, **kw):
        if not self.loaded:
            self.load_config()
        return super(LazyConfig, self).update(*args, **kw)

    def load_config(self, filename=None):
        self.loaded = True
        self.update(load(filename=filename))
        logging.config.dictConfig(self['log_config'])
        return self


#: The application configuration dictionary.
conf = LazyConfig()



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
from __future__ import absolute_import, unicode_literals

import json
import logging
import signal
import uuid

import pika
import jsonschema

from . import config
from .message import _schema_name, get_class, Message
from .signals import pre_publish_signal, publish_signal, publish_failed_signal
from .exceptions import Nack, Drop, HaltConsumer, ValidationError

_log = logging.getLogger(__name__)


class PublisherSession(object):
    """A session with blocking APIs for publishing to an AMQP broker."""

    def __init__(self, amqp_url=None, exchange=None, confirms=True):
        self._exchange = exchange or config.conf['publish_exchange']
        self._parameters = pika.URLParameters(config.conf['amqp_url'])
        if self._parameters.client_properties is None:
            self._parameters.client_properties = config.conf['client_properties']
        self._confirms = confirms
        # TODO break this out to handle automatically reconnecting and failing over, errors, etc.
        self._connection = pika.BlockingConnection(self._parameters)
        self._channel = self._connection.channel()
        if self._confirms:
            self._channel.confirm_delivery()

    def publish(self, message, exchange=None):
        """
        Publish a :class:`fedora_messaging.message.Message` to an `exchange`_ on
        the message broker.

        This is a blocking API that will retry a configurable number of times to
        publish the message.


        >>> from fedora_messaging import session, message
        >>> msg = message.Message(topic='test', body={'test':'message'})
        >>> sess = session.BlockingSession()
        >>> sess.publish(msg)

        Args:
            message (message.Message): The message to publish.
            exchange (str): The name of the AMQP exchange to publish to; defaults
                to :ref:`conf-publish-exchange`
        Raises:
            exceptions.PublishError: If publishing failed.

        .. _exchange: https://www.rabbitmq.com/tutorials/amqp-concepts.html#exchanges
        """
        pre_publish_signal.send(self, message=message)

        # Consumers use this to determine what schema to use and if they're out of date
        message.headers['fedora_messaging_schema'] = _schema_name(message.__class__)
        message.headers['fedora_messaging_schema_version'] = message.schema_version

        # Since the message can be mutated by signal listeners, validate just before sending
        message.validate()

        properties = pika.BasicProperties(
            content_type='application/json', content_encoding='utf-8', delivery_mode=2,
            headers=message.headers, message_id=str(uuid.uuid4()))
        try:
            self._channel.publish(exchange or self._exchange, message.topic.encode('utf-8'),
                                  json.dumps(message.body).encode('utf-8'), properties)
            publish_signal.send(self, message=message)
            # TODO actual error handling
        except pika.exceptions.NackError as e:
            _log.error('Message got nacked!')
            publish_failed_signal.send(self, message=message, error=e)
        except pika.exceptions.UnroutableError as e:
            _log.error('Message is unroutable!')
            publish_failed_signal.send(self, message=e, error=e)
        except pika.exceptions.ConnectionClosed as e:
            _log.warning('Connection closed to %s; attempting to reconnect...',
                         self._parameters.host)
            self._connection = pika.BlockingConnection(self._parameters)
            self._channel = self._connection.channel()
            if self._confirms:
                self._channel.confirm_delivery()
            _log.info('Successfully opened connection to %s', self._parameters.host)
            self._channel = self._connection.channel()
            self._channel.publish(exchange or self._exchange, message.topic.encode('utf-8'),
                                  json.dumps(message.body).encode('utf-8'), properties)
            publish_signal.send(self, message=message)


class ConsumerSession(object):
    """A session using the asynchronous APIs offered by Pika."""

    def __init__(self, retries=-1, retry_max_interval=60):
        self._parameters = pika.URLParameters(config.conf['amqp_url'])
        if self._parameters.client_properties is None:
            self._parameters.client_properties = config.conf['client_properties']
        self._connection = None
        self._channel = None
        self._bindings = {}
        self._retries = retries
        self._max_retry_interval = retry_max_interval
        self._retry_interval = 3
        self._retries_left = self._retries
        self._current_retry_interval = self._retry_interval
        self._running = False

        def _signal_handler(signum, frame):
            """
            Signal handler that gracefully shuts down the consumer

            Args:
                signum (int): The signal this process received.
                frame (frame): The current stack frame (unused).
            """
            if signum in (signal.SIGTERM, signal.SIGINT):
                self._shutdown()

        signal.signal(signal.SIGTERM, _signal_handler)
        signal.signal(signal.SIGINT, _signal_handler)

    def _shutdown(self):
        """Gracefully shut down the consumer and exit."""
        if self._channel:
            _log.info('Halting %r consumer sessions', self._channel.consumer_tags)
        self._running = False
        self._connection.close()

    def _on_cancelok(self, cancel_frame):
        """
        Called when the server acknowledges a cancel request.

        Args:
            cancel_frame (pika.spec.Basic.CancelOk): The cancelok frame from
                the server.
        """
        _log.info('Consumer canceled; returning all unprocessed messages to the queue')
        self._channel.basic_nack(delivery_tag=0, multiple=True, requeue=True)

    def _on_channel_open(self, channel):
        """
        Callback used when a channel is opened.

        This registers all the channel callbacks.

        Args:
            channel (pika.channel.Channel): The channel that successfully opened.
        """
        channel.add_on_close_callback(self._on_channel_close)
        channel.add_on_cancel_callback(self._on_cancel)

        channel.basic_qos(self._on_qosok, **config.conf['qos'])

    def _on_qosok(self, qosok_frame):
        """
        Callback invoked when the server acknowledges the QoS settings.

        Asserts or creates the exchanges and queues exist.

        Args:
            qosok_frame (pika.spec.Basic.Qos): The frame send from the server.
        """
        for name, args in self._exchanges.items():
            self._channel.exchange_declare(
                self._on_exchange_declareok, name, exchange_type=args['type'],
                durable=args['durable'], auto_delete=args['auto_delete'],
                arguments=args['arguments'])
        for name, args in self._queues.items():
            self._channel.queue_declare(
                self._on_queue_declareok, queue=name, durable=args['durable'],
                auto_delete=args['auto_delete'], exclusive=args['exclusive'],
                arguments=args['arguments'])

    def _on_channel_close(self, channel, reply_code, reply_text):
        """
        Callback invoked when the channel is closed.

        Args:
            channel (pika.channel.Channel): The channel that got closed.
            reply_code (int): The AMQP code indicating why the channel was closed.
            reply_text (str): The human-readable reason for the channel's closure.
        """
        _log.info('Channel %r closed (%d): %s', channel, reply_code, reply_text)
        self._channel = None

    def _on_connection_open(self, connection):
        """
        Callback invoked when the connection is successfully established.

        Args:
            connection (pika.connection.SelectConnection): The newly-estabilished
                connection.
        """
        _log.info('Successfully opened connection to %s', connection.params.host)
        self._current_retries = self._retries
        self._current_retry_interval = self._retry_interval
        self._channel = connection.channel(on_open_callback=self._on_channel_open)

    def _on_connection_close(self, connection, reply_code, reply_text):
        """
        Callback invoked when a previously-opened connection is closed.

        Args:
            connection (pika.connection.SelectConnection): The connection that
                was just closed.
            reply_code (int): The AMQP code indicating why the connection was closed.
            reply_text (str): The human-readable reason the connection was closed.
        """
        self._channel = None
        if reply_code == 200:
            # Normal shutdown, exit the consumer.
            _log.info('Server connection closed (%s), shutting down', reply_text)
        else:
            _log.warning('Connection to %s closed unexpectedly (%d): %s',
                         connection.params.host, reply_code, reply_text)
            self._reconnect()

    def _on_connection_error(self, connection, error_message):
        """
        Callback invoked when the connection failed to be established.

        Args:
            connection (pika.connection.SelectConnection): The connection that
                failed to open.
            error_message (str): The reason the connection couldn't be opened.
        """
        self._channel = None
        _log.error(error_message)
        self._reconnect()

    def _reconnect(self):
        """
        Reconnect to the broker, a configurable number of times, with a backoff.

        The connection object has a reconnect setting with an interval, but this
        backs off up to a configurable limit.
        """
        if self._retries_left != 0:
            _log.info('Reconnecting in %d seconds', self._current_retry_interval)
            self._connection.add_timeout(
                self._current_retry_interval, self._connection.connect)
            self._retries_left -= 1
            self._current_retry_interval *= 2
            if self._current_retry_interval > self._max_retry_interval:
                self._current_retry_interval = self._max_retry_interval

    def _on_exchange_declareok(self, declare_frame):
        """
        Callback invoked when an exchange is successfully declared.

        It will declare the queues in the bindings dictionary with the
        :meth:`_on_queue_declareok` callback.

        Args:
            frame (pika.spec.Exchange.DeclareOk): The DeclareOk frame from the
                server.
        """
        _log.info('Exchange declared successfully')

    def _on_queue_declareok(self, frame):
        """
        Callback invoked when a queue is successfully declared.

        Args:
            frame (pika.frame.Method): The message sent from the server.
        """
        _log.info('Successfully declared the %s queue', frame.method.queue)
        for binding in self._bindings:
            if binding['queue'] == frame.method.queue:
                for key in binding['routing_keys']:
                    _log.info('Asserting %s is bound to %s with the %s key',
                              binding['queue'], binding['exchange'], key)
                    self._channel.queue_bind(None, binding['queue'],
                                             binding['exchange'], key)
                self._channel.basic_consume(self._on_message, frame.method.queue)

    def _on_cancel(self, cancel_frame):
        """
        Callback used when the server sends a consumer cancel frame.

        Args:
            cancel_frame (pika.spec.Basic.Cancel): The cancel frame from
                the server.
        """
        _log.info('Server canceled consumer')

    def consume(self, callback, bindings=None, queues=None, exchanges=None):
        """
        Consume messages from a message queue.

        Simply define a callable to be used as the callback when messages are
        delivered and specify the queue bindings. This call blocks. The callback
        signature should accept a single positional argument which is an
        instance of a :class:`Message` (or a sub-class of it).

        Args:
            callback (callable): The callable to pass the message to when one
                arrives.
            bindings (dict): A dictionary of bindings for queues. Refer to the
                :ref:`conf-bindings` configuration documentation for the format.
            queues (dict): A dictionary of queues to ensure exist. Refer to the
                :ref:`conf-queues` configuration documentation for the format.
            exchanges (dict): A dictionary of exchanges to ensure exist. Refer
                to the :ref:`conf-exchanges` configuration documentation for the
                format.

        Raises:
            HaltConsumer: Raised when the consumer halts.
            ValidationError: When a message fails schema validation.
        """
        self._connection = pika.SelectConnection(
            self._parameters,
            on_open_callback=self._on_connection_open,
            on_open_error_callback=self._on_connection_error,
            on_close_callback=self._on_connection_close,
            stop_ioloop_on_close=True,
        )
        self._bindings = bindings or config.conf['bindings']
        self._queues = queues or config.conf['queues']
        self._exchanges = exchanges or config.conf['exchanges']
        self._consumer_callback = callback
        self._running = True
        while self._running:
            self._connection.ioloop.start()

    def _on_message(self, channel, delivery_frame, properties, body):
        """
        Callback when a message is received from the server.

        This method wraps a user-registered callback for message delivery. It
        decodes the message body, determines the message schema to validate the
        message with, and validates the message before passing it on to the user
        callback.

        This also handles acking, nacking, and rejecting messages based on
        exceptions raised by the consumer callback. For detailed documentation
        on the user-provided callback, see the user guide on consuming.

        Args:
            channel (pika.channel.Channel): The channel from which the message
                was received.
            delivery_frame (pika.spec.Deliver): The delivery frame which includes
                details about the message like content encoding and its delivery
                tag.
            properties (pika.spec.BasicProperties): The message properties like
                the message headers.
            body (bytes): The message payload.

        Raises:
            HaltConsumer: Raised when the consumer halts.
            ValidationError: When a message fails schema validation.
        """
        _log.debug('Message arrived with delivery tag %s', delivery_frame.delivery_tag)
        try:
            MessageClass = get_class(properties.headers['fedora_messaging_schema'])
        except KeyError:
            _log.error('Message (headers=%r, body=%r) arrived without a schema header.'
                       ' A publisher is misbehaving!', properties.headers, body)
            MessageClass = Message

        if properties.content_encoding is None:
            _log.error('Message arrived without a content encoding')
            properties.content_encoding = 'utf-8'
        try:
            body = body.decode(properties.content_encoding)
        except UnicodeDecodeError:
            _log.error('Unable to decode message body %r with %s content encoding',
                       body, delivery_frame.content_encoding)

        try:
            body = json.loads(body)
        except ValueError as e:
            _log.error('Failed to load message body %r, %r', body, e)
            raise ValidationError(e)

        message = MessageClass(
            body=body, headers=properties.headers, topic=delivery_frame.routing_key)
        try:
            message.validate()
            _log.debug('Successfully validated message %r', message)
        except jsonschema.exceptions.ValidationError as e:
            _log.error('Message validation of %r failed: %r', message, e)
            raise ValidationError(e)

        try:
            _log.info('Consuming message from topic "%s" (id %s)', message.topic,
                      properties.message_id)
            self._consumer_callback(message)
            channel.basic_ack(delivery_tag=delivery_frame.delivery_tag)
        except Nack:
            _log.info('Returning message id %s to the queue', properties.message_id)
            channel.basic_nack(delivery_tag=delivery_frame.delivery_tag, requeue=True)
        except Drop:
            _log.info('Dropping message id %s', properties.message_id)
            channel.basic_nack(delivery_tag=delivery_frame.delivery_tag, requeue=False)
        except HaltConsumer as e:
            _log.info('Consumer requested halt, returning messages to queue')
            channel.basic_nack(delivery_tag=delivery_frame.delivery_tag, requeue=True)
            self._shutdown()
            if e.exit_code != 0:
                raise
        except Exception as e:
            _log.exception("Received unexpected exception from consumer callback")
            channel.basic_nack(delivery_tag=0, multiple=True, requeue=True)
            self._shutdown()
            raise HaltConsumer(exit_code=1, reason=e)



"""Exceptions raised by Fedora Messaging."""


class BaseException(Exception):
    """The base class for all exceptions raised by fedora_messaging."""


class PublishException(BaseException):
    """Base class for exceptions related to publishing."""


class PublishReturned(PublishException):
    """Raised when the broker rejects and returns the message to the publisher."""


class ConnectionError(BaseException):
    """Raised if a general connection error occurred."""


class ConsumeException(BaseException):
    """Base class for exceptions related to consuming."""


class Nack(ConsumeException):
    """
    Consumer callbacks should raise this to indicate they wish the message they
    are currently processing to be re-queued.
    """


class Drop(ConsumeException):
    """
    Consumer callbacks should raise this to indicate they wish the message they
    are currently processing to be dropped.
    """


class HaltConsumer(ConsumeException):
    """
    Consumer callbacks should raise this exception if they wish the consumer to
    be shut down.

    Args:
        exit_code (int): The exit code to use when halting.
    """

    def __init__(self, exit_code=0, reason=None, **kwargs):
        super(HaltConsumer, self).__init__(**kwargs)
        self.exit_code = exit_code
        self.reason = reason


class ValidationError(BaseException):
    """This error is raised when a message fails validation."""



import os

from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst')) as fd:
    README = fd.read()


def get_requirements(requirements_file='requirements.txt'):
    """Get the contents of a file listing the requirements.

    Args:
        requirements_file (str): The path to the requirements file, relative
                                 to this file.

    Returns:
        list: the list of requirements, or an empty list if
              ``requirements_file`` could not be opened or read.
    """
    with open(requirements_file) as fd:
        lines = fd.readlines()
    dependencies = []
    for line in lines:
        maybe_dep = line.strip()
        if maybe_dep.startswith('#'):
            # Skip pure comment lines
            continue
        if maybe_dep.startswith('git+'):
            # VCS reference for dev purposes, expect a trailing comment
            # with the normal requirement
            __, __, maybe_dep = maybe_dep.rpartition('#')
        else:
            # Ignore any trailing comment
            maybe_dep, __, __ = maybe_dep.partition('#')
        # Remove any whitespace and assume non-empty results are dependencies
        maybe_dep = maybe_dep.strip()
        if maybe_dep:
            dependencies.append(maybe_dep)
    return dependencies


setup(
    name='fedora_messaging',
    version='0.1.0a2',
    description="A set of tools for using Fedora's messaging infrastructure",
    long_description=README,
    # Possible options are at https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 1 - Planning',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    license='GPLv2+',
    maintainer='Fedora Infrastructure Team',
    maintainer_email='infrastructure@lists.fedoraproject.org',
    platforms=['Fedora', 'GNU/Linux'],
    keywords='fedora',
    packages=find_packages(exclude=('fedora_messaging.tests',
                                    'fedora_messaging.tests.*')),
    include_package_data=True,
    zip_safe=False,
    install_requires=get_requirements(),
    tests_require=get_requirements(requirements_file='dev-requirements.txt'),
    test_suite='fedora_messaging.tests',
    entry_points={
        'console_scripts': [
            'fedora-messaging=fedora_messaging.cli:cli',
        ],
        'fedora.messages': ['base.message=fedora_messaging.message:Message']
    }
)



"""The API for publishing messages and consuming from message queues."""

from . import _session
from .signals import pre_publish_signal, publish_signal, publish_failed_signal
from .message import Message


__all__ = (
    'Message',
    'consume',
    'publish',
    'pre_publish_signal',
    'publish_signal',
    'publish_failed_signal',
)

_session_cache = None


def consume(callback, bindings=None):
    """
    Start a message consumer that executes the provided callback when messages are
    received.

    This API is blocking and will not return until the process receives a signal
    from the operating system.

    The callback receives a single positional argument, the message:

    >>> from fedora_messaging import api
    >>> def my_callback(message):
    ...     print(message)
    >>> bindings = [{'exchange': 'amq.topic', 'queue': 'demo', 'routing_keys': ['#']}]
    >>> api.consume(my_callback, bindings=bindings)

    For complete documentation on writing consumers, see the :ref:`consumers`
    documentation.

    Args:
        callback (callable): A callable object that accepts one positional argument,
            a :class:`Message`.
        bindings (dict or list of dict): The bindings to use when consuming. This
            should be the same format as the :ref:`conf-bindings` configuration.
    """
    if isinstance(bindings, dict):
        bindings = [bindings]
    session = _session.ConsumerSession()
    session.consume(callback, bindings)


def publish(message, exchange=None):
    """
    Publish a message to an exchange.

    This is a synchronous call, meaning that when this function returns, an
    acknowledgment has been received from the message broker and you can be
    certain the message was published successfully.

    There are some cases where an error occurs despite your message being
    successfully published. For example, if a network partition occurs after
    the message is received by the broker. Therefore, you may publish duplicate
    messages. For complete details, see the :ref:`publishing` documentation.

    >>> from fedora_messaging import api
    >>> message = api.Message(body={'Hello': 'world'}, topic='Hi')
    >>> api.publish(message)

    Args:
        message (message.Message): The message to publish.
        exchange (str): The name of the AMQP exchange to publish to; defaults to
            :ref:`conf-publish-exchange`

    Raises:
        fedora_messaging.exceptions.PublishReturned: Raised if the broker rejects the
            message.
        fedora_messaging.exceptions.ConnectionError: Raised if a connection error
            occurred before the publish confirmation arrived.

    # TODO doc retry behavior, when messages could get double published, etc.
    """
    # TODO make thread-local registry, probably
    global _session_cache
    if _session_cache is None:
        _session_cache = _session.PublisherSession(exchange)
    _session_cache.publish(message, exchange=exchange)



"""
This module defines the base class of message objects and keeps a registry of
known message implementations. This registry is populated from Python entry
points. The registry keys are Python paths and the values are sub-classes of
:class:`Message`.

When publishing, the API will add a header with the schema used so the consumer
can locate the correct schema.

To implement your own message schema, simply create a class that inherits the
:class:`Message` class, and add an entry point in your Python package under the
"fedora.messages" group. For example, an entry point for the :class:`Message`
schema would be::

    entry_points = {
        'fedora.messages': ['base.message=fedora_messaging.message:Message']
    }

Since every client needs to have the message schema installed, you should define
this class in a small Python package of its own. Note that at this time, the
entry point name is unused.
"""

import json
import logging

import jsonschema
import pkg_resources


_log = logging.getLogger(__name__)

# Maps regular expressions to message classes
_class_registry = {}

# Used to load the registry automatically on first use
_registry_loaded = False


def get_class(schema_name):
    """
    Retrieve the message class associated with the schema name.

    If no match is found, the default schema is returned and a warning is logged.

    Args:
        schema_name (six.text_type): The name of the :class:`Message` sub-class;
            this is typically the Python path.

    Returns:
        Message: A sub-class of :class:`Message` to create the message from.
    """
    global _registry_loaded
    if not _registry_loaded:
        load_message_classes()
        _registry_loaded = True

    try:
        return _class_registry[schema_name]
    except KeyError:
        _log.warning('The schema "%s" is not in the schema registry! Either install '
                     'the package with its schema definition or define a schema. '
                     'Falling back to the default schema...', schema_name)
        return _class_registry[_schema_name(Message)]


def load_message_classes():
    """Load the 'fedora.messages' entry points and register the message classes."""
    for message in pkg_resources.iter_entry_points('fedora.messages'):
        cls = message.load()
        _log.info("Registering the '%r' class as a Fedora Message", cls)
        _class_registry[_schema_name(cls)] = cls


def _schema_name(cls):
    """
    Get the Python path of a class, used to identify the message schema.

    Args:
        cls (object): The class (not the instance of the class).

    Returns:
        str: The path in the format "<module>:<class_name>".
    """
    return '{}:{}'.format(cls.__module__, cls.__name__)


class Message(object):
    """
    Messages are simply JSON-encoded objects. This allows message authors to
    define a schema and implement Python methods to abstract the raw message
    from the user. This allows the schema to change and evolve without breaking
    the user-facing API.

    A message class includes a topic. This topic is used by message consumers
    to filter what messages they receive. Topics should be a string of words
    separated by '.' characters, with a length limit of 255 bytes. Because of
    this limit, it is best to avoid non-ASCII characters as well as
    variable-length components where the total size of the topic would exceed
    255 bytes.

    Attributes:
        topic (six.text_type): The message topic as a unicode string.
        headers_schema (dict): A `JSON schema <http://json-schema.org/>`_ to be used with
            :func:`jsonschema.validate` to validate the message headers.
        body_schema (dict): A `JSON schema <http://json-schema.org/>`_ to be used with
            :func:`jsonschema.validate` to validate the message headers.

    Args:
        headers (dict): A set of message headers. Consult the headers schema for
            expected keys and values.
        body (dict): The message body. Consult the body schema for expected keys
            and values.
        topic (six.text_type): The message topic as a unicode string. If this is
            not provided, the default topic for the class is used.
    """

    topic = ''
    headers_schema = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'description': 'Schema for message headers',
        'type': 'object',
    }
    body_schema = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'description': 'Schema for message body',
        'type': 'object',
    }
    schema_version = 1

    def __init__(self, body=None, headers=None, topic=None):
        self.headers = headers or {}
        self.body = body or {}
        if topic:
            self.topic = topic

    def __str__(self):
        """
        A human-readable representation of this message.

        This should provide a detailed representation of the message, much like the body
        of an email.

        The default implementation is to format the raw message topic, headers, and body.
        Applications use this to present messages to users.
        """
        return 'Topic: {t}\n\nHeaders: {h}\n\nBody: {b}'.format(
            t=self.topic,
            h=json.dumps(self.headers, sort_keys=True, indent=4),
            b=json.dumps(self.body, sort_keys=True, indent=4)
        )

    def __repr__(self):
        """
        Provide a printable representation of the object that can be passed to func:`eval`.
        """
        return "{}(body={}, headers={}, topic={})".format(
            self.__class__.__name__, repr(self.body), repr(self.headers), repr(self.topic))

    def __eq__(self, other):
        """
        Two messages of the same class with the same topic, headers, and body are equal.

        Args:
            other (object): The object to check for equality.

        Returns:
            bool: True if the messages are equal.
        """
        return (isinstance(other, self.__class__) and self.topic == other.topic and
                self.body == other.body and self.headers == other.headers)

    def summary(self):
        """
        A short, human-readable representation of this message.

        This should provide a short summary of the message, much like the subject line
        of an email.

        The default implementation is to simply return the message topic.
        """
        return self.topic

    def validate(self):
        """
        Validate the headers and body with the message schema, if any.

        Raises:
            jsonschema.ValidationError: If either the message headers or the message body
                are invalid.
            jsonschema.SchemaError: If either the message header schema or the message body
                schema are invalid.
        """
        _log.debug('Validating message headers "%r" with schema "%r", version %d',
                   self.headers, self.headers_schema, self.schema_version)
        jsonschema.validate(self.headers, self.headers_schema)
        _log.debug('Validating message body "%r" with schema "%r", version %d',
                   self.headers, self.headers_schema, self.schema_version)
        jsonschema.validate(self.body, self.body_schema)
